package com.kakao.sdk.auth

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.net.Uri
import android.os.Bundle
import android.os.Handler
import android.os.Looper
import android.os.ResultReceiver
import com.kakao.sdk.common.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.*
import com.kakao.sdk.common.util.IntentResolveClient
import io.reactivex.Single
import io.reactivex.SingleEmitter
import java.net.HttpURLConnection

/**
 * OAuth 2.0 Spec 의 authorization code 를 발급받기 위해 사용되는 클라이언트.
 *
 * @author kevin.kang. Created on 2018. 3. 20..
 */
class RxAuthCodeClient(
    private val authApiClient: RxAuthApiClient = AuthApiClient.rx,
    private val intentResolveClient: IntentResolveClient = IntentResolveClient.instance
) {

    /**
     * 카카오톡 간편 로그인을 통하여 authorization code 를 발급 받는다.
     */
    @JvmOverloads
    fun authorizeWithTalk(
        context: Context,
        requestCode: Int,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.clientId}://oauth",
        clientId: String = KakaoSdk.applicationContextInfo.clientId,
        approvalType: String = "individual",
        kaHeader: String = KakaoSdk.applicationContextInfo.kaHeader
    ): Single<String> {
        return Single.create {

            val talkIntent = talkLoginIntent(clientId, approvalType, redirectUri, kaHeader)
            val resolvedIntent = intentResolveClient.resolveTalkIntent(context, talkIntent)
            if (resolvedIntent == null) {
                it.onError(
                    ClientError(
                        ClientErrorCause.NotSupported,
                        "KakaoTalk not installed"
                    )
                )
                return@create
            }
            context.startActivity(
                Intent(context, TalkAuthCodeActivity::class.java)
                    .putExtra(
                        com.kakao.sdk.auth.Constants.KEY_LOGIN_INTENT,
                        intentResolveClient.resolveTalkIntent(context, talkIntent)
                    )
                    .putExtra(Constants.KEY_REQUEST_CODE, requestCode)
                    .putExtra(Constants.KEY_RESULT_RECEIVER, resultReceiver(it))
            )
        }
    }

    /**
     * Custom Tabs (기본 브라우저) 를 사용하여 authorization code 를 발급 받는다.
     *
     * @param context instance for starting an activity
     * @param clientId native app key
     *
     * @return [Single] that will emit authorization code
     */
    @JvmOverloads
    fun authorize(
        context: Context,
        clientId: String = KakaoSdk.applicationContextInfo.clientId,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.clientId}://oauth",
        approvalType: String = "individual"
    ): Single<String> {
        return Single.create {
            val fullUri = UriUtility.authorizeUri(
                clientId = clientId,
                redirectUri = redirectUri,
                approvalType = approvalType
            )
            context.startActivity(authCodeIntent(context, fullUri, redirectUri, resultReceiver(it)))
        }
    }

    /**
     * 유저의 동의항목을 업데이트할 수 있는 authorization code 를 발급받는다.
     *
     * @param context [Context] instance for starting an activity
     * @param scopes list of scopes to be updated
     *
     * @return [Single] that will emit authorization code
     */
    @JvmOverloads
    fun authorizeWithNewScopes(
        context: Context,
        scopes: List<String>,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.clientId}://oauth",
        clientId: String = KakaoSdk.applicationContextInfo.clientId,
        approvalType: String = "individual"
    ): Single<String> {
        return authApiClient.agt(clientId).flatMap {
            Single.create<String> { emitter ->
                val uri = UriUtility.authorizeUri(
                    clientId = clientId,
                    agt = it.agt,
                    redirectUri = redirectUri,
                    scopes = scopes,
                    approvalType = approvalType
                )
                val intent = authCodeIntent(context, uri, redirectUri, resultReceiver(emitter))
                context.startActivity(intent)
            }
        }
    }

    @JvmSynthetic
    internal fun talkLoginIntent(
        clientId: String,
        approvalType: String,
        redirectUri: String,
        kaHeader: String
    ): Intent {
        return Intent().setAction("com.kakao.talk.intent.action.CAPRI_LOGGED_IN_ACTIVITY")
            .addCategory(Intent.CATEGORY_DEFAULT)
            .putExtra(Constants.EXTRA_APPLICATION_KEY, clientId)
            .putExtra(Constants.EXTRA_REDIRECT_URI, redirectUri)
            .putExtra(Constants.EXTRA_KA_HEADER, kaHeader)
            .putExtra(Constants.APPROVAL_TYPE, approvalType)
    }

    @JvmSynthetic
    internal fun authCodeIntent(
        context: Context,
        fullUri: Uri,
        redirectUri: String,
        resultReceiver: ResultReceiver
    ) = Intent(context, AuthCodeCustomTabsActivity::class.java)
        .putExtra(Constants.KEY_RESULT_RECEIVER, resultReceiver)
        .putExtra(Constants.KEY_FULL_URI, fullUri)
        .putExtra(Constants.KEY_REDIRECT_URI, redirectUri)

    @JvmSynthetic
    internal fun resultReceiver(emitter: SingleEmitter<String>): ResultReceiver {
        return object : ResultReceiver(Handler(Looper.getMainLooper())) {
            override fun onReceiveResult(resultCode: Int, resultData: Bundle?) {
                this@RxAuthCodeClient.onReceivedResult(resultCode, resultData, emitter)
            }
        }
    }

    @JvmSynthetic
    internal fun onReceivedResult(
        resultCode: Int,
        resultData: Bundle?,
        emitter: SingleEmitter<String>
    ) {
        when (resultCode) {
            Activity.RESULT_OK -> {
                val uri = resultData?.getParcelable<Uri>(Constants.KEY_URL)
                val code = uri?.getQueryParameter(Constants.CODE)
                if (code != null) {
                    emitter.onSuccess(code)
                    return
                }
                val error = uri?.getQueryParameter(Constants.ERROR) as String
                val errorDescription = uri.getQueryParameter(Constants.ERROR_DESCRIPTION) as String
                emitter.onError(
                    OAuthError(
                        HttpURLConnection.HTTP_MOVED_TEMP,
                        KakaoJson.fromJson(
                            error,
                            AuthErrorCause::class.java
                        )
                            ?: AuthErrorCause.Unknown,
                        AuthErrorResponse(error, errorDescription)
                    )
                )
            }
            Activity.RESULT_CANCELED -> {
                val exception =
                    resultData?.getSerializable(Constants.KEY_EXCEPTION)
                            as KakaoSdkError
                emitter.onError(exception)
            }
            else -> throw IllegalArgumentException("Unknown resultCode in RxAuthCodeClient#onReceivedResult()")
        }
    }
}

val AuthCodeClient.Companion.rx by lazy { RxAuthCodeClient() }